@@ -8,12 +8,11 @@ pub(crate) use self::core::BLOB_INDEX_FILE_EXTENSION;
 pub(crate) use self::core::{Blob, FileName};
 pub use self::entry::Entry;
 pub(crate) use self::file::File;
-pub use self::index::bloom as filter;
-pub(crate) use self::index::{BloomConfig, IndexConfig};
+pub(crate) use self::index::IndexConfig;
 pub(crate) use super::prelude::*;
 
 mod prelude {
     pub(crate) use super::*;
-    pub(crate) use index::{FilterResult, Index};
+    pub(crate) use index::Index;
     pub(crate) use std::sync::atomic::AtomicU64;
 }
