@@ -5,9 +5,6 @@ use futures::stream::FuturesOrdered;
 use tokio::fs::{create_dir, create_dir_all};
 
 const BLOB_FILE_EXTENSION: &str = "blob";
-const LOCK_FILE: &str = "pearl.lock";
-
-const O_EXCL: i32 = 128;
 
 /// A main storage struct.
 ///
@@ -35,37 +32,24 @@ const O_EXCL: i32 = 128;
 /// ```
 ///
 /// [`Key`]: trait.Key.html
-#[derive(Debug)]
+#[derive(Debug, Clone)]
 pub struct Storage<K: Key> {
-    pub(crate) inner: Inner,
-    observer: Observer,
-    marker: PhantomData<K>,
+    pub(crate) inner: Inner<K>,
+    observer: Observer<K>,
 }
 
 #[derive(Debug, Clone)]
-pub(crate) struct Inner {
+pub(crate) struct Inner<K: Key> {
     pub(crate) config: Config,
-    pub(crate) safe: Arc<RwLock<Safe>>,
+    pub(crate) safe: Arc<RwLock<Safe<K>>>,
     next_blob_id: Arc<AtomicUsize>,
     pub(crate) ioring: Option<Rio>,
 }
 
 #[derive(Debug)]
-pub(crate) struct Safe {
-    pub(crate) active_blob: Option<Box<Blob>>,
-    pub(crate) blobs: Arc<RwLock<Vec<Blob>>>,
-    lock_file: Option<StdFile>,
-}
-
-impl<K: Key> Clone for Storage<K> {
-    #[must_use]
-    fn clone(&self) -> Self {
-        Self {
-            inner: self.inner.clone(),
-            observer: self.observer.clone(),
-            marker: PhantomData,
-        }
-    }
+pub(crate) struct Safe<K: Key> {
+    pub(crate) active_blob: Option<Box<Blob<K>>>,
+    pub(crate) blobs: Arc<RwLock<HierarchicalFilters<K, Bloom, Blob<K>>>>,
 }
 
 async fn work_dir_content(wd: &Path) -> Result<Option<Vec<DirEntry>>> {
@@ -91,16 +75,12 @@ async fn work_dir_content(wd: &Path) -> Result<Option<Vec<DirEntry>>> {
     Ok(content)
 }
 
-impl<K: Key> Storage<K> {
+impl<K: Key + 'static> Storage<K> {
     pub(crate) fn new(config: Config, ioring: Option<Rio>) -> Self {
         let dump_sem = config.dump_sem();
         let inner = Inner::new(config, ioring);
         let observer = Observer::new(inner.clone(), dump_sem);
-        Self {
-            inner,
-            observer,
-            marker: PhantomData,
-        }
+        Self { inner, observer }
     }
 
     /// [`init()`] used to prepare all environment to further work.
@@ -203,14 +183,17 @@ impl<K: Key> Storage<K> {
     /// Fails if there are some errors during dump
     /// [`close_active_blob_in_background()`]: struct.Storage.html#method.create_active_blob_async
     pub async fn try_close_active_blob(&self) -> Result<()> {
-        self.inner.close_active_blob().await
+        let result = self.inner.close_active_blob().await;
+        self.observer.try_dump_old_blob_indexes().await;
+        result
     }
 
     /// Dumps active blob
     /// NOTICE! This function returns immediately, so you can't check result of operation. If you
     /// want be sure about operation's result, use [`try_close_active_blob()`]
     pub async fn close_active_blob_in_background(&self) {
-        self.observer.close_active_blob().await
+        self.observer.close_active_blob().await;
+        self.observer.try_dump_old_blob_indexes().await
     }
 
     /// Sets last blob from closed blobs as active if there is no active blobs
@@ -249,8 +232,7 @@ impl<K: Key> Storage<K> {
     ///
     /// [`write_with`]: Storage::write_with
     pub async fn write(&self, key: impl AsRef<K>, value: Vec<u8>) -> Result<()> {
-        self.write_with_optional_meta(key.as_ref(), value, None)
-            .await
+        self.write_with_optional_meta(key, value, None).await
     }
 
     /// Similar to [`write`] but with metadata
@@ -267,25 +249,27 @@ impl<K: Key> Storage<K> {
     /// # Errors
     /// Fails if duplicates are not allowed and record already exists.
     pub async fn write_with(&self, key: impl AsRef<K>, value: Vec<u8>, meta: Meta) -> Result<()> {
-        self.write_with_optional_meta(key.as_ref(), value, Some(meta))
-            .await
+        self.write_with_optional_meta(key, value, Some(meta)).await
     }
 
     async fn write_with_optional_meta(
         &self,
-        key: &K,
+        key: impl AsRef<K>,
         value: Vec<u8>,
         meta: Option<Meta>,
     ) -> Result<()> {
+        let key = key.as_ref();
         debug!("storage write with {:?}, {}b, {:?}", key, value.len(), meta);
         // if active blob is set, this function will only check this fact and return false
         if self.try_create_active_blob().await.is_ok() {
             info!("Active blob was set during write operation");
         }
-        if !self.inner.config.allow_duplicates()
-            && self.contains_with(key.as_ref(), meta.as_ref()).await?
-        {
-            warn!("record with key {:?} and meta {:?} exists", key, meta);
+        if !self.inner.config.allow_duplicates() && self.contains_with(key, meta.as_ref()).await? {
+            warn!(
+                "record with key {:?} and meta {:?} exists",
+                key.as_ref(),
+                meta
+            );
             return Ok(());
         }
         let record = Record::create(key, value, meta.unwrap_or_default())
@@ -324,8 +308,9 @@ impl<K: Key> Storage<K> {
     /// [`read_with`]: Storage::read_with
     #[inline]
     pub async fn read(&self, key: impl AsRef<K>) -> Result<Vec<u8>> {
-        debug!("storage read {:?}", key.as_ref());
-        self.read_with_optional_meta(key.as_ref(), None).await
+        let key = key.as_ref();
+        debug!("storage read {:?}", key);
+        self.read_with_optional_meta(key, None).await
     }
     /// Reads data matching given key and metadata
     /// # Examples
@@ -354,7 +339,7 @@ impl<K: Key> Storage<K> {
     /// # Errors
     /// Fails after any disk IO errors.
     pub async fn read_all(&self, key: impl AsRef<K>) -> Result<Vec<Entry>> {
-        let key = key.as_ref().as_ref();
+        let key = key.as_ref();
         let mut all_entries = Vec::new();
         let safe = self.inner.safe.read().await;
         let active_blob = safe
@@ -370,8 +355,8 @@ impl<K: Key> Storage<K> {
         }
         let blobs = safe.blobs.read().await;
         let entries_closed_blobs = blobs
-            .iter()
-            .map(|b| b.read_all_entries(key))
+            .iter_possible_childs(key)
+            .map(|b| b.1.data.read_all_entries(key))
             .collect::<FuturesUnordered<_>>();
         entries_closed_blobs
             .try_filter_map(future::ok)
@@ -388,7 +373,6 @@ impl<K: Key> Storage<K> {
     async fn read_with_optional_meta(&self, key: &K, meta: Option<&Meta>) -> Result<Vec<u8>> {
         debug!("storage read with optional meta {:?}, {:?}", key, meta);
         let safe = self.inner.safe.read().await;
-        let key = key.as_ref();
         if let Some(ablob) = safe.active_blob.as_ref() {
             match ablob.read_any(key, meta, true).await {
                 Ok(data) => {
@@ -401,29 +385,30 @@ impl<K: Key> Storage<K> {
         Self::get_any_data(&safe, key, meta).await
     }
 
-    async fn get_data_last(safe: &Safe, key: &[u8], meta: Option<&Meta>) -> Result<Vec<u8>> {
+    async fn get_data_last(safe: &Safe<K>, key: &K, meta: Option<&Meta>) -> Result<Vec<u8>> {
         let blobs = safe.blobs.read().await;
-        let blobs_stream: FuturesOrdered<_> = blobs
-            .iter()
-            .enumerate()
-            .map(|(i, blob)| async move {
-                if blob.check_filters_non_blocking(key).await {
-                    Some(i)
+        let possible_blobs = blobs
+            .iter_possible_childs_rev(key)
+            .map(|(id, blob)| async move {
+                if matches!(blob.data.check_filters(key).await, Ok(true)) {
+                    Some(id)
                 } else {
                     None
                 }
             })
-            .collect();
-        let possible_blobs: Vec<_> = blobs_stream.filter_map(|e| e).collect().await;
+            .collect::<FuturesOrdered<_>>()
+            .filter_map(|x| x)
+            .collect::<Vec<_>>()
+            .await;
         debug!(
             "len of possible blobs: {} (start len: {})",
             possible_blobs.len(),
             blobs.len()
         );
         let stream: FuturesOrdered<_> = possible_blobs
-            .iter()
-            .rev()
-            .map(|i| blobs[*i].read_any(key, meta, false))
+            .into_iter()
+            .filter_map(|id| blobs.get_child(id))
+            .map(|blob| blob.data.read_any(key, meta, false))
             .collect();
         debug!("read with optional meta {} closed blobs", stream.len());
         let mut task = stream.skip_while(Result::is_err);
@@ -431,11 +416,11 @@ impl<K: Key> Storage<K> {
     }
 
     #[allow(dead_code)]
-    async fn get_data_any(safe: &Safe, key: &[u8], meta: Option<&Meta>) -> Result<Vec<u8>> {
+    async fn get_data_any(safe: &Safe<K>, key: &K, meta: Option<&Meta>) -> Result<Vec<u8>> {
         let blobs = safe.blobs.read().await;
         let stream: FuturesUnordered<_> = blobs
-            .iter()
-            .map(|blob| blob.read_any(key, meta, true))
+            .iter_possible_childs_rev(key)
+            .map(|blob| blob.1.data.read_any(key, meta, true))
             .collect();
         debug!("read with optional meta {} closed blobs", stream.len());
         let mut task = stream.skip_while(Result::is_err);
@@ -445,7 +430,7 @@ impl<K: Key> Storage<K> {
             .with_context(|| "no results in closed blobs")
     }
 
-    async fn get_any_data(safe: &Safe, key: &[u8], meta: Option<&Meta>) -> Result<Vec<u8>> {
+    async fn get_any_data(safe: &Safe<K>, key: &K, meta: Option<&Meta>) -> Result<Vec<u8>> {
         Self::get_data_last(safe, key, meta).await
     }
 
@@ -464,14 +449,6 @@ impl<K: Key> Storage<K> {
                     .with_context(|| format!("blob {} dump failed", blob.name())),
             )
         }
-        safe.lock_file = None;
-        if let Some(work_dir) = self.inner.config.work_dir() {
-            res = res.and(
-                std::fs::remove_file(work_dir.join(LOCK_FILE))
-                    .map(|_| info!("lock released"))
-                    .with_context(|| "failed to remove lockfile ({:?})"),
-            );
-        }
         res
     }
 
@@ -533,27 +510,6 @@ impl<K: Key> Storage<K> {
             ))
             .with_context(|| "failed to prepare work dir");
         }
-        self.try_lock_dir(path)
-            .await
-            .with_context(|| format!("failed to lock work dir: {}", path.display()))
-    }
-
-    async fn try_lock_dir<'a>(&'a self, path: &'a Path) -> Result<()> {
-        let lock_file_path = path.join(LOCK_FILE);
-        debug!("try to open lock file: {}", lock_file_path.display());
-        let lock_file = StdOpenOptions::new()
-            .create(true)
-            .write(true)
-            .custom_flags(O_EXCL)
-            .open(&lock_file_path)
-            .map_err(|e| {
-                error!("working directory is locked: {:?}", lock_file_path);
-                error!("check if any other bob instances are running");
-                error!("or delete .lock file and try again");
-                e
-            })?;
-        debug!("{} not locked", path.display());
-        self.inner.safe.write().await.lock_file = Some(lock_file);
         Ok(())
     }
 
@@ -595,14 +551,16 @@ impl<K: Key> Storage<K> {
 
         let mut safe = self.inner.safe.write().await;
         safe.active_blob = active_blob;
-        *safe.blobs.write().await = blobs;
+        *safe.blobs.write().await =
+            HierarchicalFilters::from_vec(self.inner.config.bloom_filter_group_size(), 1, blobs)
+                .await;
         self.inner
             .next_blob_id
             .store(safe.max_id().await.map_or(0, |i| i + 1), ORD);
         Ok(())
     }
 
-    async fn pop_active(blobs: &mut Vec<Blob>, config: &Config) -> Result<Box<Blob>> {
+    async fn pop_active(blobs: &mut Vec<Blob<K>>, config: &Config) -> Result<Box<Blob<K>>> {
         let mut active_blob = blobs
             .pop()
             .ok_or_else(|| {
@@ -611,7 +569,7 @@ impl<K: Key> Storage<K> {
                 Error::from(ErrorKind::Uninitialized)
             })?
             .boxed();
-        active_blob.load_index(K::LEN).await?;
+        active_blob.load_index().await?;
         Ok(active_blob)
     }
 
@@ -620,7 +578,7 @@ impl<K: Key> Storage<K> {
         ioring: Option<Rio>,
         disk_access_sem: Arc<Semaphore>,
         config: &Config,
-    ) -> Result<Vec<Blob>> {
+    ) -> Result<Vec<Blob<K>>> {
         debug!("read working directory content");
         let dir_content = files.iter().map(DirEntry::path);
         debug!("read {} entities", dir_content.len());
@@ -638,7 +596,7 @@ impl<K: Key> Storage<K> {
             .map(|file| async {
                 let sem = disk_access_sem.clone();
                 let _sem = sem.acquire().await.expect("sem is closed");
-                Blob::from_file(file.clone(), ioring.clone(), config.index(), K::LEN)
+                Blob::from_file(file.clone(), ioring.clone(), config.index())
                     .await
                     .map_err(|e| (e, file))
             })
@@ -734,12 +692,11 @@ impl<K: Key> Storage<K> {
     /// `contains` returns either "definitely in storage" or "definitely not".
     /// # Errors
     /// Fails because of any IO errors
-    pub async fn contains(&self, key: K) -> Result<bool> {
-        let key = key.as_ref();
-        self.contains_with(key, None).await
+    pub async fn contains(&self, key: impl AsRef<K>) -> Result<bool> {
+        self.contains_with(key.as_ref(), None).await
     }
 
-    async fn contains_with(&self, key: &[u8], meta: Option<&Meta>) -> Result<bool> {
+    async fn contains_with(&self, key: &K, meta: Option<&Meta>) -> Result<bool> {
         let inner = self.inner.safe.read().await;
         if let Some(active_blob) = &inner.active_blob {
             if active_blob.contains(key, meta).await? {
@@ -747,8 +704,8 @@ impl<K: Key> Storage<K> {
             }
         }
         let blobs = inner.blobs.read().await;
-        for blob in blobs.iter() {
-            if blob.contains(key, meta).await? {
+        for blob in blobs.iter_possible_childs(key) {
+            if blob.1.data.contains(key, meta).await? {
                 return Ok(true);
             }
         }
@@ -763,19 +720,45 @@ impl<K: Key> Storage<K> {
     /// In other words, `check_filters` returns either "possibly in storage" or "definitely not".
     pub async fn check_filters(&self, key: impl AsRef<K>) -> Option<bool> {
         let key = key.as_ref();
-        trace!("[{:?}] check in blobs bloom filter", &key.to_vec());
+        trace!("[{:?}] check in blobs bloom filter", key);
         let inner = self.inner.safe.read().await;
-        let in_active = inner
-            .active_blob
-            .as_ref()
-            .map_or(false, |active_blob| active_blob.check_filters(key.as_ref()));
-        let in_closed = inner
-            .blobs
-            .read()
-            .await
+        let in_active = if let Some(active_blob) = inner.active_blob.as_ref() {
+            active_blob
+                .check_filters(key)
+                .await
+                .map_err(|e| {
+                    error!(
+                        "Failed to check filter for active blob for key {:?}: {}",
+                        key, e
+                    );
+                    e
+                })
+                .unwrap_or(false)
+        } else {
+            false
+        };
+        if in_active {
+            return Some(true);
+        }
+
+        let blobs = inner.blobs.read().await;
+        let (offloaded, in_memory): (Vec<&Blob<K>>, Vec<&Blob<K>>) =
+            blobs.iter().partition(|blob| blob.is_filter_offloaded());
+
+        let in_closed = in_memory
             .iter()
-            .any(|blob| blob.check_filters(key.as_ref()) == true);
-        Some(in_active || in_closed)
+            .any(|blob| blob.check_filters_in_memory(key));
+        if in_closed {
+            return Some(true);
+        }
+
+        let in_closed_offloaded = offloaded
+            .iter()
+            .map(|blob| blob.check_filters(key))
+            .collect::<FuturesUnordered<_>>()
+            .any(|value| value.unwrap_or(false))
+            .await;
+        Some(in_closed_offloaded)
     }
 
     /// Total records count in storage.
@@ -810,7 +793,8 @@ impl<K: Key> Storage<K> {
     /// Or if there are some problems with syncronization.
     /// [`close_active_blob_in_background()`]: struct.Storage.html#method.close_active_blob_async
     pub async fn force_update_active_blob(&self, predicate: ActiveBlobPred) {
-        self.observer.force_update_active_blob(predicate).await
+        self.observer.force_update_active_blob(predicate).await;
+        self.observer.try_dump_old_blob_indexes().await
     }
 
     fn launch_observer(&mut self) {
@@ -818,11 +802,11 @@ impl<K: Key> Storage<K> {
     }
 }
 
-impl Inner {
+impl<K: Key + 'static> Inner<K> {
     fn new(config: Config, ioring: Option<Rio>) -> Self {
         Self {
+            safe: Arc::new(RwLock::new(Safe::new(config.bloom_filter_group_size()))),
             config,
-            safe: Arc::new(RwLock::new(Safe::new())),
             next_blob_id: Arc::new(AtomicUsize::new(0)),
             ioring,
         }
@@ -872,12 +856,10 @@ impl Inner {
         if safe.active_blob.is_none() {
             Err(Error::active_blob_doesnt_exist().into())
         } else {
-            // FIXME: write lock is still held, so everyone will wait for dump, maybe it's better
-            // to derive this operation to `try_dump_old_blob_indexes`
-            safe.active_blob.as_mut().unwrap(/*None case checked*/).dump().await?;
             // always true
             if let Some(ablob) = safe.active_blob.take() {
-                safe.blobs.write().await.push(*ablob);
+                ablob.fsyncdata().await?;
+                safe.blobs.write().await.push(*ablob).await;
             }
             Ok(())
         }
@@ -947,23 +929,22 @@ impl Inner {
         self.safe.read().await.fsyncdata().await
     }
 
-    pub(crate) async fn try_dump_old_blob_indexes(&mut self, sem: Arc<Semaphore>) {
+    pub(crate) async fn try_dump_old_blob_indexes(&self, sem: Arc<Semaphore>) {
         self.safe.write().await.try_dump_old_blob_indexes(sem).await;
     }
 }
 
-impl Safe {
-    fn new() -> Self {
+impl<K: Key + 'static> Safe<K> {
+    fn new(group_size: usize) -> Self {
         Self {
             active_blob: None,
-            blobs: Arc::new(RwLock::new(Vec::new())),
-            lock_file: None,
+            blobs: Arc::new(RwLock::new(HierarchicalFilters::new(group_size, 1))),
         }
     }
 
     async fn max_id(&self) -> Option<usize> {
         let active_blob_id = self.active_blob.as_ref().map(|blob| blob.id());
-        let blobs_max_id = self.blobs.read().await.last().map(Blob::id);
+        let blobs_max_id = self.blobs.read().await.last().map(|x| x.id());
         active_blob_id.max(blobs_max_id)
     }
 
@@ -996,10 +977,10 @@ impl Safe {
         Ok(())
     }
 
-    pub(crate) async fn replace_active_blob(&mut self, blob: Box<Blob>) -> Result<()> {
+    pub(crate) async fn replace_active_blob(&mut self, blob: Box<Blob<K>>) -> Result<()> {
         let old_active = self.active_blob.replace(blob);
         if let Some(blob) = old_active {
-            self.blobs.write().await.push(*blob);
+            self.blobs.write().await.push(*blob).await;
         }
         Ok(())
     }
@@ -1024,7 +1005,7 @@ impl Safe {
 }
 
 /// Trait `Key`
-pub trait Key: AsRef<[u8]> + Debug {
+pub trait Key: AsRef<[u8]> + Debug + Clone + Send + Sync + Ord + From<Vec<u8>> + Default {
     /// Key must have fixed length
     const LEN: u16;
 
@@ -1033,3 +1014,66 @@ pub trait Key: AsRef<[u8]> + Debug {
         self.as_ref().to_vec()
     }
 }
+
+#[async_trait::async_trait]
+impl<K: Key + 'static> BloomProvider<K> for Storage<K> {
+    type Filter = <Blob<K> as BloomProvider<K>>::Filter;
+    async fn check_filter(&self, item: &K) -> FilterResult {
+        let inner = self.inner.safe.read().await;
+        let active = inner
+            .active_blob
+            .as_ref()
+            .map(|b| b.check_filter_fast(item))
+            .unwrap_or_default();
+        let ret = inner.blobs.read().await.check_filter(item).await;
+        ret + active
+    }
+
+    fn check_filter_fast(&self, _item: &K) -> FilterResult {
+        FilterResult::NeedAdditionalCheck
+    }
+
+    async fn offload_buffer(&mut self, needed_memory: usize, level: usize) -> usize {
+        let inner = self.inner.safe.read().await;
+        let ret = inner
+            .blobs
+            .write()
+            .await
+            .offload_buffer(needed_memory, level)
+            .await;
+        ret
+    }
+
+    async fn get_filter(&self) -> Option<Self::Filter> {
+        let inner = self.inner.safe.read().await;
+        let mut ret = inner.blobs.read().await.get_filter_fast().cloned();
+        if let Some(filter) = &mut ret {
+            if let Some(active_filter) = inner
+                .active_blob
+                .as_ref()
+                .map(|b| b.get_filter_fast())
+                .flatten()
+            {
+                if !filter.checked_add_assign(active_filter) {
+                    return None;
+                }
+            }
+        }
+        ret
+    }
+
+    fn get_filter_fast(&self) -> Option<&Self::Filter> {
+        None
+    }
+
+    async fn filter_memory_allocated(&self) -> usize {
+        let safe = self.inner.safe.read().await;
+        let active = safe
+            .active_blob
+            .as_ref()
+            .map_or(0, |blob| blob.filter_memory_allocated());
+
+        let closed = safe.blobs.read().await.filter_memory_allocated().await;
+        active + closed
+    }
+}
